package com.sun.org.apache.xml.internal.serializer;

import java.util.Iterator;
import java.util.Enumeration;
import com.sun.org.apache.xml.internal.serializer.utils.WrappedRuntimeException;
import java.util.StringTokenizer;
import java.util.Properties;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import com.sun.org.apache.xalan.internal.utils.SecuritySupport;
import java.nio.charset.UnsupportedCharsetException;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.Charset;
import java.io.UnsupportedEncodingException;
import java.io.BufferedWriter;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.io.OutputStream;

public final class Encodings
{
    private static final int m_defaultLastPrintable = 127;
    private static final String ENCODINGS_FILE = "com/sun/org/apache/xml/internal/serializer/Encodings.properties";
    private static final String ENCODINGS_PROP = "com.sun.org.apache.xalan.internal.serialize.encodings";
    static final String DEFAULT_MIME_ENCODING = "UTF-8";
    private static final Encodings.EncodingInfos _encodingInfos;
    
    public Encodings() {
    }
    
    static Writer getWriter(OutputStream output, String encoding) throws UnsupportedEncodingException {
        EncodingInfo ei = Encodings._encodingInfos.findEncoding(toUpperCaseFast(encoding));
        if (ei != null) {
            try {
                return new BufferedWriter(new OutputStreamWriter(output, ei.javaName));
            }
            catch (UnsupportedEncodingException ex) {}
        }
        return new BufferedWriter(new OutputStreamWriter(output, encoding));
    }
    
    public static int getLastPrintable() {
        return 127;
    }
    
    static EncodingInfo getEncodingInfo(String encoding) {
        String normalizedEncoding = toUpperCaseFast(encoding);
        EncodingInfo ei = Encodings._encodingInfos.findEncoding(normalizedEncoding);
        if (ei == null) {
            try {
                Charset c = Charset.forName(encoding);
                String name = c.name();
                ei = new EncodingInfo(name, name);
                Encodings._encodingInfos.putEncoding(normalizedEncoding, ei);
            }
            catch (IllegalCharsetNameException | UnsupportedCharsetException x) {
                ei = new EncodingInfo(null, null);
            }
        }
        return ei;
    }
    
    public static boolean isRecognizedEncoding(String encoding) {
        String normalizedEncoding = toUpperCaseFast(encoding);
        EncodingInfo ei = Encodings._encodingInfos.findEncoding(normalizedEncoding);
        return ei != null;
    }
    
    private static String toUpperCaseFast(String s) {
        boolean different = false;
        int mx = s.length();
        char[] chars = new char[mx];
        char ch;
        for (int i = 0; i < mx; chars[i] = ch, ++i) {
            ch = s.charAt(i);
            if ('a' <= ch && ch <= 'z') {
                ch -= 32;
                different = true;
            }
        }
        if (different) {
            String upper = String.valueOf(chars);
            return upper;
        }
        String upper = s;
        return upper;
    }
    
    static String getMimeEncoding(String encoding) {
        if (null == encoding) {
            try {
                encoding = SecuritySupport.getSystemProperty("file.encoding", "UTF8");
                if (null != encoding) {
                    String jencoding = (encoding.equalsIgnoreCase("Cp1252") || encoding.equalsIgnoreCase("ISO8859_1") || encoding.equalsIgnoreCase("8859_1") || encoding.equalsIgnoreCase("UTF8")) ? "UTF-8" : convertJava2MimeEncoding(encoding);
                    encoding = ((null != jencoding) ? jencoding : "UTF-8");
                }
                else {
                    encoding = "UTF-8";
                }
            }
            catch (SecurityException se) {
                encoding = "UTF-8";
            }
        }
        else {
            encoding = convertJava2MimeEncoding(encoding);
        }
        return encoding;
    }
    
    private static String convertJava2MimeEncoding(String encoding) {
        EncodingInfo enc = Encodings._encodingInfos.getEncodingFromJavaKey(toUpperCaseFast(encoding));
        if (null != enc) {
            return enc.name;
        }
        return encoding;
    }
    
    public static String convertMime2JavaEncoding(String encoding) {
        EncodingInfo info = Encodings._encodingInfos.findEncoding(toUpperCaseFast(encoding));
        return (info != null) ? info.javaName : encoding;
    }
    
    static boolean isHighUTF16Surrogate(char ch) {
        return '?' <= ch && ch <= '?';
    }
    
    static boolean isLowUTF16Surrogate(char ch) {
        return '?' <= ch && ch <= '?';
    }
    
    static int toCodePoint(char highSurrogate, char lowSurrogate) {
        int codePoint = (highSurrogate - 55296 << 10) + (lowSurrogate - 56320) + 65536;
        return codePoint;
    }
    
    static int toCodePoint(char ch) {
        int codePoint = ch;
        return codePoint;
    }
    
    static String access$000(String x0) {
        return toUpperCaseFast(x0);
    }
    
    static {
        _encodingInfos = new Encodings.EncodingInfos();
    }
    
    private static final class EncodingInfos
    {
        private final Map<String, EncodingInfo> _encodingTableKeyJava;
        private final Map<String, EncodingInfo> _encodingTableKeyMime;
        private final Map<String, EncodingInfo> _encodingDynamicTable;
        
        private EncodingInfos() {
            this._encodingTableKeyJava = new HashMap();
            this._encodingTableKeyMime = new HashMap();
            this._encodingDynamicTable = Collections.synchronizedMap(new HashMap());
            this.loadEncodingInfo();
        }
        
        private InputStream openEncodingsFileStream() throws MalformedURLException, IOException {
            String urlString = null;
            InputStream is = null;
            try {
                urlString = SecuritySupport.getSystemProperty("com.sun.org.apache.xalan.internal.serialize.encodings", "");
            }
            catch (SecurityException ex) {}
            if (urlString != null && urlString.length() > 0) {
                URL url = new URL(urlString);
                is = url.openStream();
            }
            if (is == null) {
                is = SecuritySupport.getResourceAsStream("com/sun/org/apache/xml/internal/serializer/Encodings.properties");
                return is;
            }
            return is;
        }
        
        private Properties loadProperties() throws MalformedURLException, IOException {
            Properties props = new Properties();
            try (final InputStream is = this.openEncodingsFileStream()) {
                if (is != null) {
                    props.load(is);
                }
            }
            return props;
        }
        
        private String[] parseMimeTypes(String val) {
            int pos = val.indexOf(32);
            if (pos < 0) {
                return new String[] { val };
            }
            StringTokenizer st = new StringTokenizer(val.substring(0, pos), ",");
            String[] values = new String[st.countTokens()];
            int i = 0;
            while (st.hasMoreTokens()) {
                values[i] = st.nextToken();
                ++i;
            }
            return values;
        }
        
        private String findCharsetNameFor(String name) {
            try {
                return Charset.forName(name).name();
            }
            catch (Exception x) {
                return null;
            }
        }
        
        private String findCharsetNameFor(String javaName, String[] mimes) {
            String cs = this.findCharsetNameFor(javaName);
            if (cs != null) {
                return javaName;
            }
            int length = mimes.length;
            int i = 0;
            while (i < length) {
                String m = mimes[i];
                cs = this.findCharsetNameFor(m);
                if (cs != null) {
                    break;
                }
                ++i;
            }
            return cs;
        }
        
        private void loadEncodingInfo() {
            try {
                Properties props = this.loadProperties();
                Enumeration<Object> keys = props.keys();
                Map<String, EncodingInfo> canonicals = new HashMap();
                while (keys.hasMoreElements()) {
                    String javaName = (String)keys.nextElement();
                    String[] mimes = this.parseMimeTypes(props.getProperty(javaName));
                    String charsetName = this.findCharsetNameFor(javaName, mimes);
                    if (charsetName != null) {
                        String kj = Encodings.access$000(javaName);
                        String kc = Encodings.access$000(charsetName);
                        for (int i = 0; i < mimes.length; ++i) {
                            String mimeName = mimes[i];
                            String km = Encodings.access$000(mimeName);
                            EncodingInfo info = new EncodingInfo(mimeName, charsetName);
                            this._encodingTableKeyMime.put(km, info);
                            if (!canonicals.containsKey(kc)) {
                                canonicals.put(kc, info);
                                this._encodingTableKeyJava.put(kc, info);
                            }
                            this._encodingTableKeyJava.put(kj, info);
                        }
                    }
                    else {
                        continue;
                    }
                }
                for (Map.Entry<String, EncodingInfo> e : this._encodingTableKeyJava.entrySet()) {
                    e.setValue(canonicals.get(Encodings.access$000(((EncodingInfo)e.getValue()).javaName)));
                }
            }
            catch (MalformedURLException mue) {
                throw new WrappedRuntimeException(mue);
            }
            catch (IOException ioe) {
                throw new WrappedRuntimeException(ioe);
            }
        }
        
        EncodingInfo findEncoding(String normalizedEncoding) {
            EncodingInfo info = (EncodingInfo)this._encodingTableKeyJava.get(normalizedEncoding);
            if (info == null) {
                info = (EncodingInfo)this._encodingTableKeyMime.get(normalizedEncoding);
            }
            if (info == null) {
                info = (EncodingInfo)this._encodingDynamicTable.get(normalizedEncoding);
                return info;
            }
            return info;
        }
        
        EncodingInfo getEncodingFromMimeKey(String normalizedMimeName) {
            return (EncodingInfo)this._encodingTableKeyMime.get(normalizedMimeName);
        }
        
        EncodingInfo getEncodingFromJavaKey(String normalizedJavaName) {
            return (EncodingInfo)this._encodingTableKeyJava.get(normalizedJavaName);
        }
        
        void putEncoding(String key, EncodingInfo info) {
            this._encodingDynamicTable.put(key, info);
        }
    }
}
